# Powershell based synchronization of aqua with LDAP and Active Directory.
# This script is compatible with aqua 18.0 or later.
#
# This script synchronizes the users in the LDAP direcotry with aqua.
# Users which do not yet exist in aqua are created and a default license and project
# is assigned to them.
# For existing users, the first name, last name and email address are updated. Project
# and license assignments are not modified.
# The script uses the aqua API and needs administrator credentials
#
# WARNING: Due to limitations of the API, the script only detects existing users when the
# user name has the exact same capitalization. Otherwise, the script thinks that the user
# is new.
#
# Version history:
# 2017-10-12 Version 1:
#   - Synchronization with LDAP
# 2018-02-28 Version 2:
#   - Synchronization with Active Directory
#   - Deactivation of users based on a flag
# 2019-11-25 Version 3:
#   - Deactivate aqua users which are missing in the LDAP result
#   - Improved testing functionality

Import-Module "${PSScriptRoot}\DirectoryFunctionsLDAP.psm1" -Force
Import-Module "${PSScriptRoot}\Logging.psm1" -Force

. "${PSScriptRoot}\Configuration.ps1"

$global:logfile = "${PSScriptRoot}\logs\UserSync.log"

Add-Type -AssemblyName "System.Web"

$userStatusActive = 0
$userStatusDeactivated = 1
$userStatusLicenseDeactivated = 2

$knownUsersFile = "${PSScriptRoot}\data\knownusers.txt"

function Login {
    $global:client = New-WebServiceProxy -Uri ("$webserviceUrl" + "?wsdl") -Namespace aqua.WebServiceProxy
    $global:session = $client.Login($apiUser, $apiPassword)
    LogInfo "Login to aqua API successful"
}

function Logout {
    $client.Logout($global:session)
    LogInfo "Logout from aqua API successful"
}

function ReadFileAsSet {
    [CmdletBinding()]
    param (
        [Parameter(Mandatory = $true)]
        [string]
        $Path
    )

    $content = Get-Content $Path
    if ($content -eq $null) {
        $content = @()
    }

    return New-Object System.Collections.Generic.HashSet[string] @(,[string[]]$content)
}


### Main ###
try {
    if (-not (Test-Path $knownUsersFile)) {
        if (-not (Test-Path "${PSScriptRoot}\data\")) {
            New-Item -ItemType Directory -Path "${PSScriptRoot}\data\" | Out-Null
        }
        New-Item -ItemType File -Path $knownUsersFile | Out-Null
    }

    $knownUsers = ReadFileAsSet -Path $knownUsersFile
    $seenUsers = New-Object System.Collections.Generic.HashSet[string]

    LogInfo "$(Get-Date -Format g): Starting synchronization of users in directory service"
    LogInfo "Going to fetch users from the directory..."
    
    $directoryUsers = GetDirectoryUsers

    # This line can be uncommented to reduce the user list for testing
    # $directoryUsers = $directoryUsers | Where { $_.Username -eq "SampleUser" }

    LogInfo "$(($directoryUsers | Measure-Object).Count) users fetched from directory"
    LogInfo "Going to synchronize users with aqua..."

    Login

    $defaultProjectId = $null
    $defaultRoleId = $null

    if ((-not [System.String]::IsNullOrEmpty($defaultProjectName)) -and (-not [System.String]::IsNullOrEmpty($defaultRoleName))) {
        $projects = $client.GetVisibleProjects($global:session);
        $project = $projects | Where-Object { $_.ProjectName -eq $defaultProjectName } | Select-Object -First 1
        if ($project -ne $null) {
            $defaultProjectId = $project.ProjectId
        }

        $roles = $client.GetRoles($global:session);
        $role = $roles | Where-Object { $_.Name -eq $defaultRoleName } | Select-Object -First 1
        if ($role -ne $null) {
            $defaultRoleId = $role.Id
        }
    }

    foreach ($directoryUser in $directoryUsers) {
        if ($directoryUser.Username -eq $null) {
            LogError "Skipping user $($directoryUser.DistinguishedName). No username found"
            continue
        }

        if (-not $knownUsers.Contains($directoryUser.Username)) {
            Add-Content -Path $knownUsersFile -Value $directoryUser.Username
        }
        $seenUsers.Add($directoryUser.Username) | Out-Null
        
        $aquaUsers = $client.GetUsers($global:session) | where { $_.Username -ieq $directoryUser.Username }
        
        if ($aquaUsers -eq $null -or $aquaUsers.length -eq 0) {
            if ($directoryUser.Disabled -eq $true) {
                LogInfo "User $($directoryUser.Username) not in aqua, but disabled currently in the directory so will not be created in aqua."
            } else {
                LogInfo "New directory user $($directoryUser.Username) found"
                $userPassword = [System.Web.Security.Membership]::GeneratePassword(16,0)
                $dummy = $client.CreateUserWithFloatingFlag($global:session, $directoryUser.UserName, $userPassword, $directoryUser.Email, $directoryUser.FirstName, $directoryUser.LastName, $floatingLicense, $licenseProfileCode);
                if (($defaultProjectId -ne $null) -and ($defaultRoleId -ne $null)) {
                    $aquaUser = $client.GetUser($global:session, $directoryUser.Username)
                    if ($aquaUser -ne $null) {
                        $assignment = New-Object -TypeName aqua.WebServiceProxy.UserRoleAssignment
                        $assignment.UserId = $aquaUser.Id
                        $assignment.RoleId = $defaultRoleId
                        $assignment.FolderId = 0 
                        $client.SaveUserAssignmentsForProject($global:session, $defaultProjectId, @($assignment), $false)
                    } else {
                        LogError "Could not find user $($directoryUser.Username) after creating or updating it in aqua. Check the capitalization of the user name."
                    }
                }

                LogInfo "Successfully created user $($directoryUser.Username)"
            }
        } elseif ($aquaUsers.length -eq 1) {
                $aquaUser = $aquaUsers[0]
                if (($directoryUser.Disabled -eq $true) -and ($aquaUser.UserStatus -eq  $userStatusActive) -and (-not ($global:userWhiteList | Where-Object { $_ -ieq $aquaUser.Username }))) {
                    LogInfo "Directory user $($aquaUser.Username) is disabled/deactived. Deactivating also in aqua..."
                    $client.SetUserActivationStatus($global:session, $aquaUser.Id, $userStatusDeactivated)
                    LogInfo "aqua user $($aquaUser.Username) sucessfully deactivated"
                } else {
                    $changes = $false
                    $changes = $changes -or ($directoryUser.FirstName -ne $aquaUser.FirstName)
                    $changes = $changes -or ($directoryUser.LastName -ne $aquaUser.LastName)
                    $changes = $changes -or ($directoryUser.Email -ne $aquaUser.Email)

                    if ($changes) {
                        LogInfo "Changes detected for user $($aquaUser.Username)"
                        $licenseProfiles = $client.GetUserProfileAssignments($global:session, $aquaUser.Id)
                        $dummy = $client.CreateUserWithFloatingFlag($global:session, $directoryUser.Username, $null, $directoryUser.Email, $directoryUser.FirstName, $directoryUser.LastName, $floatingLicense, $licenseProfileCode);
                        $client.SaveUserProfileAssignments($global:session, $aquaUser.Id, $licenseProfiles)
                        LogInfo "Successfully updated user $($aquaUser.Username)"
                    } 
                }
        } else {
            LogError "Skipping user $($directoryUser.Username). The user is not unique in aqua. The user name exists multiple times in different capitalizations in aqua."
        }
    }

    if ($global:deactivateKnownMissingUsers) {
        LogInfo "Going to disable known missing users in aqua..."
        $aquaUsers = $client.GetUsers($global:session)

        foreach ($aquaUser in $aquaUsers) {
            if ($aquaUser.UserStatus -ne $userStatusDeactivated) {
                if (($knownUsers | Where-Object { $_ -ieq $aquaUser.Username }) -and (-not ($seenUsers | Where-Object { $_ -ieq $aquaUser.Username })) -and (-not ($global:userWhiteList | Where-Object { $_ -ieq $aquaUser.Username }))) {
                    LogInfo "aqua user $($aquaUser.Username) is missing in LDAP results. Deactivating the user in aqua..."
                    $client.SetUserActivationStatus($global:session, $aquaUser.Id, $userStatusDeactivated)
                    LogInfo "aqua user $($aquaUser.Username) sucessfully deactivated"
                }
            }
        }
    }

    if ($global:deactivateAllMissingUsers) {
        LogInfo "Going to disable all missing users in aqua which are not whitelisted..."
        $aquaUsers = $client.GetUsers($global:session)

        foreach ($aquaUser in $aquaUsers) {
            if ($aquaUser.UserStatus -ne $userStatusDeactivated) {
                if ((-not ($seenUsers | Where-Object { $_ -ieq $aquaUser.Username })) -and (-not ($global:userWhiteList | Where-Object { $_ -ieq $aquaUser.Username }))) {
                    LogInfo "aqua user $($aquaUser.Username) is missing in LDAP results. Deactivating the user in aqua..."
                    $client.SetUserActivationStatus($global:session, $aquaUser.Id, $userStatusDeactivated)
                    LogInfo "aqua user $($aquaUser.Username) sucessfully deactivated"
                }
            }
        }
    }

    Logout
    LogInfo "Synchronization finished with return code 0"
    return 0
} catch {
    LogError $_.Exception
    return -1
}